package cli

import (
	"flag"
	"io"
	"log"
	"os"
	"time"

	"github.com/it-novum/nsta/nstaclient"
	"github.com/it-novum/nsta/transfer"
	"gopkg.in/ini.v1"
)

// Cli holds the information about command line parameters
type Cli struct {
	configPath    string
	Configuration *IniConfig
}

type IniConfig struct {
	Gearman             string `ini:"gearman"`
	ListenHTTP          string `ini:"listen-http"`
	ListenHTTPS         string `ini:"listen-https"`
	TLSKey              string `ini:"tls-key"`
	TLSCert             string `ini:"tls-cert"`
	Mode                string `ini:"mode"`
	APIKey              string `ini:"api-key"`
	Endpoint            string `ini:"endpoint"`
	SatelliteID         int    `ini:"satellite-id"`
	MyCNF               string `ini:"mycnf"`
	NagiosPath          string `ini:"nagios-path"`
	InsecureHTTPS       bool   `ini:"insecure-https"`
	Proxy               string `ini:"proxy"`
	LogFile             string `ini:"log-file"`
	AccessLogFile       string `ini:"access-log-file"`
	Interval            int    `ini:"interval"`
	DisableInstantPush  bool   `ini:"disable-instant-push"`
	SSHReloadTimeoutSec int    `ini:"ssh-reload-timeout-sec"`
}

func loadIniConfig(configPath string) (*IniConfig, error) {
	cfg, err := ini.Load(configPath)
	if err != nil {
		return nil, err
	}
	nstaSec, err := cfg.GetSection("nsta")
	if err != nil {
		return nil, err
	}

	iniConfig := &IniConfig{
		Gearman:             "127.0.0.1:4730",
		ListenHTTP:          "127.0.0.1:7473",
		ListenHTTPS:         ":7474",
		TLSKey:              "/etc/ssl/private/ssl-cert-snakeoil.key",
		TLSCert:             "/etc/ssl/certs/ssl-cert-snakeoil.pem",
		Mode:                "server",
		SatelliteID:         -1,
		MyCNF:               "/opt/openitc/etc/mysql/mysql.cnf",
		NagiosPath:          "/opt/openitc/nagios",
		LogFile:             "-",
		AccessLogFile:       "-",
		Interval:            5,
		SSHReloadTimeoutSec: 300,
	}

	if err := nstaSec.MapTo(iniConfig); err != nil {
		return nil, err
	}

	return iniConfig, err
}

// ParseCLI uses flag to parse the command line parameters
func parseCLI() *Cli {
	c := &Cli{}
	flag.StringVar(&c.configPath, "config", "/opt/openitc/etc/nsta/nsta.ini", "Path to nsta.ini configuration")
	flag.Parse()

	iniConfig, err := loadIniConfig(c.configPath)
	if err != nil {
		log.Fatalln("Could not load nsta configuration file: ", err)
	}

	if iniConfig.Mode == "https_pull" {
		if iniConfig.APIKey == "" {
			log.Fatalln("Verify nsta configuration failed: missing -api-key")
		}
		if iniConfig.SatelliteID == -1 {
			log.Fatalln("Verify nsta configuration failed: missing -satellite-id")
		}
	}
	if iniConfig.Mode == "https_push" {
		if iniConfig.APIKey == "" {
			log.Fatalln("Verify nsta configuration failed: missing -api-key")
		}
		if iniConfig.SatelliteID == -1 {
			log.Fatalln("Verify nsta configuration failed: missing -satellite-id")
		}
		if iniConfig.Endpoint == "" {
			log.Fatalln("Verify nsta configuration failed: missing -endpoint url")
		}
	}

	c.Configuration = iniConfig

	return c
}

func Run() {
	var err error

	cli := parseCLI()
	cfg := cli.Configuration

	accessLogWriter := io.Discard
	if cfg.AccessLogFile != "" {
		if cfg.AccessLogFile == "-" {
			accessLogWriter = os.Stdout
		} else {
			accessLogWriter, err = os.OpenFile(cfg.AccessLogFile, os.O_APPEND|os.O_CREATE, os.FileMode(0o644))
			if err != nil {
				log.Println(err)
				os.Exit(1)
			}
		}
	}

	logWriter := io.Discard
	if cfg.LogFile != "" {
		if cfg.LogFile == "-" {
			logWriter = os.Stdout
		} else {
			logWriter, err = os.OpenFile(cfg.LogFile, os.O_APPEND|os.O_CREATE, os.FileMode(0o644))
			if err != nil {
				log.Println(err)
				os.Exit(1)
			}
		}
	}

	log.SetOutput(logWriter)

	// Mode "server" is always running on the Main openITCOCKPIT Server (Master)
	if cfg.Mode == "server" {
		m := &transfer.MainSystem{
			AccessLogWriter:       accessLogWriter,
			MyCnf:                 cfg.MyCNF,
			GearmanAddress:        cfg.Gearman,
			WebserverAddressHTTP:  cfg.ListenHTTP,
			WebserverAddressHTTPS: cfg.ListenHTTPS,
			CertFile:              cfg.TLSCert,
			KeyFile:               cfg.TLSKey,
			NagiosPath:            cfg.NagiosPath,
			SSHReloadTimeoutSec:   cfg.SSHReloadTimeoutSec,
		}
		if err := m.StartForeground(); err != nil {
			log.Fatalln(err)
		}

		// "https_push" is running on a Satellite Server
	} else if cfg.Mode == "https_push" {
		c := &nstaclient.ClientPush{
			ClientSettings: nstaclient.ClientSettings{
				GearmanAddress:    cfg.Gearman,
				Endpoint:          cfg.Endpoint,
				SatelliteID:       cfg.SatelliteID,
				APIKey:            cfg.APIKey,
				NagiosPath:        cfg.NagiosPath,
				Proxy:             cfg.Proxy,
				VerifyCertificate: !cfg.InsecureHTTPS,
				PushInterval:      time.Second * time.Duration(cfg.Interval),
				InstantPush:       !cfg.DisableInstantPush,
			},
		}
		if err := c.StartForeground(); err != nil {
			log.Fatalln(err)
		}

		// "https_pull" is running on a Satellite Server
	} else if cfg.Mode == "https_pull" {
		c := &nstaclient.ClientPull{
			ClientSettings: nstaclient.ClientSettings{
				GearmanAddress:        cfg.Gearman,
				SatelliteID:           cfg.SatelliteID,
				APIKey:                cfg.APIKey,
				NagiosPath:            cfg.NagiosPath,
				AccessLogWriter:       accessLogWriter,
				WebserverAddressHTTP:  cfg.ListenHTTP,
				WebserverAddressHTTPS: cfg.ListenHTTPS,
				CertFile:              cfg.TLSCert,
				KeyFile:               cfg.TLSKey,
			},
		}
		if err := c.StartForeground(); err != nil {
			log.Fatalln(err)
		}
	}
}
