package gearman

import (
	"log"
	"time"

	jsoniter "github.com/json-iterator/go"
)

var json = jsoniter.ConfigCompatibleWithStandardLibrary

// MessageType identifes the type of a statusengine message
type MessageType int

const (
	// MessageOCHP is an OCHP statusengine message
	MessageOCHP MessageType = 0
	// MessageOCSP is an OCSP statusengine message
	MessageOCSP MessageType = 1
	// MessageCommandSatTX is a command statusengine message
	MessageCommandSatTX MessageType = 2
	// MessageCustomInput is a phpnsta custom message transmission
	MessageCustomInput MessageType = 3
	// MessageCommandMasterTX transfers commands from sat to master
	MessageCommandMasterTX MessageType = 4
	// MessageCheckmk transfers Checkmk commands from master to sat
	MessageCheckmk MessageType = 5
	// MessageTimesync signals a timesync to satellite
	MessageTimesync MessageType = 6
	// MessageOitcAgent transfers oITC Agent commands from master to sat
	MessageOitcAgent MessageType = 7
)

func (t MessageType) WorkerQueue() string {
	switch t {
	case MessageOCHP:
		return "statusngin_ochp"
	case MessageOCSP:
		return "statusngin_ocsp"
	case MessageCommandSatTX:
		return "statusngin_cmd_sattx"
	case MessageCustomInput:
		return "phpnsta_custom_input"
	case MessageCommandMasterTX:
		return "statusngin_cmd_mastertx"
	case MessageCheckmk:
		return "oitc_checkmk_sattx"
	case MessageOitcAgent:
		return "oitc_agent_sattx"
	default:
		panic("unimplemented")
	}
}

func (t MessageType) ClientQueue() string {
	switch t {
	case MessageOCHP:
		return "statusngin_worker_ochp"
	case MessageOCSP:
		return "statusngin_worker_ocsp"
	case MessageCommandSatTX:
		return "statusngin_cmd"
	case MessageCustomInput:
		return "phpnsta_custom_data"
	case MessageCommandMasterTX:
		return "statusngin_cmd"
	case MessageCheckmk:
		return "oitc_gearman"
	case MessageOitcAgent:
		return "oitc_gearman"
	default:
		panic("unimplemented")
	}
}

var MessageTypeMain = []MessageType{
	MessageCommandSatTX,
	MessageCheckmk,
	MessageOitcAgent,
}

var MessageTypeSatellite = []MessageType{
	MessageOCHP,
	MessageOCSP,
	MessageCustomInput,
	MessageCommandMasterTX,
}

// MessageContentTX represents a typical otic gearman message
type MessageContentTX struct {
	SatelliteID int
	Command     string
	Messages    []*MessageContentTX
}

// Message holds the information of a message for queuing
type Message struct {
	Type   MessageType
	String string
}

// Content parses the content of String
func (m *Message) Content() *MessageContentTX {
	c := &MessageContentTX{}
	err := json.Unmarshal([]byte(m.String), c)
	if err != nil {
		log.Println("could not parse gearman content: ", err)
		return nil
	}
	return c
}

// CheckmkDataMessage represents the information for checkmk scan
type CheckmkDataMessage struct {
	ScanID   int
	Hostuuid string
	Task     string
	File     string
	FilePath string
}

// CheckmkMessage is a special message for checkmk related commands
type CheckmkMessage struct {
	SatelliteID int
	Command     string
	Data        CheckmkDataMessage
}

// CheckmkResponse holds the information for the response to the gearman worker
type CheckmkResponse struct {
	Task             string `json:"task"`
	SatelliteID      int
	Command          string
	ScanID           int
	CheckTypesResult string
	DumpHostResult   string
	RawInfoResult    string
	Error            string
}

// TimesyncMessage represents a timesync message
type TimesyncMessage struct {
	SatelliteID int
	SetTime     time.Time
}

// Serialize returns the php representation of the response message, required by gearman_worker
func (c *CheckmkResponse) Serialize() (string, error) {
	c.Task = "CheckmkSatResult"
	r, err := json.Marshal(c)
	if err != nil {
		return "", err
	}
	return string(r), nil
}

// OitcAgentMessage is the message send by openITCOCKPIT to trigger the oITC Agent
type OitcAgentMessage struct {
	SatelliteID int
	Command     string
	Data        OitcAgentDataMessage
}

// OitcAgentDataMessage holds the configuration to query the agent
type OitcAgentDataMessage struct {
	TaskID             int
	AgentConfigId      int
	Task               string
	Address            string
	Port               int
	UseAutossl         bool
	AutosslSuccessful  bool
	TestConnectionOnly bool //Only test the connection and exchange Auto-TLS certificates if needed
	UseHttps           bool
	Insecure           bool
	BasicAuth          bool
	Username           string
	Password           string
	PushAgentId        int    // Used for assign_host task
	HostUuid           string // Used for assign_host and query_push task
}

// OitcAgentResponseMessage is a special message for openITCOCKPIT Agent related commands
type OitcAgentResponseMessage struct {
	SatelliteID int
	Command     string
	Data        CheckmkDataMessage
}

// OitcAgentResponse holds the information for the response to the gearman worker
type OitcAgentResponse struct {
	Task                      string `json:"task"`
	AgentConfigId             int
	SatelliteID               int
	Command                   string
	TaskID                    int
	RawResult                 string
	Error                     string
	IsGetAllPushAgentsRequest bool
}

// SerializeOitcAgent returns the php representation of the response message, required by gearman_worker
func (c *OitcAgentResponse) SerializeOitcAgent() (string, error) {
	c.Task = "OitcAgentSatResult"
	r, err := json.Marshal(c)
	if err != nil {
		return "", err
	}
	return string(r), nil
}
